@@ -1,4 +1,6 @@
-/* $Id: handler.c,v 1.13.4.4 2005/07/14 12:00:36 vanhu Exp $ */
+/*	$NetBSD: handler.c,v 1.9.6.6 2007/06/06 09:20:12 vanhu Exp $	*/
+
+/* Id: handler.c,v 1.28 2006/05/26 12:17:29 manubsd Exp */
 
 /*
  * Copyright (C) 1995, 1996, 1997, and 1998 WIDE Project.
@@ -48,6 +50,10 @@
 #include "sockmisc.h"
 #include "debug.h"
 
+#ifdef ENABLE_HYBRID
+#include <resolv.h>
+#endif
+
 #include "schedule.h"
 #include "grabmyaddr.h"
 #include "algorithm.h"
@@ -69,6 +75,8 @@
 #include "gcmalloc.h"
 #include "nattraversal.h"
 
+#include "sainfo.h"
+
 #ifdef HAVE_GSSAPI
 #include "gssapi.h"
 #endif
@@ -108,6 +116,7 @@ getph1byindex(index)
 	return NULL;
 }
 
+
 /*
  * search for isakmp handler by i_ck in index.
  */
@@ -138,14 +147,24 @@ getph1byaddr(local, remote)
 {
 	struct ph1handle *p;
 
+	plog(LLV_DEBUG2, LOCATION, NULL, "getph1byaddr: start\n");
+	plog(LLV_DEBUG2, LOCATION, NULL, "local: %s\n", saddr2str(local));
+	plog(LLV_DEBUG2, LOCATION, NULL, "remote: %s\n", saddr2str(remote));
+
 	LIST_FOREACH(p, &ph1tree, chain) {
 		if (p->status == PHASE1ST_EXPIRED)
 			continue;
+		plog(LLV_DEBUG2, LOCATION, NULL, "p->local: %s\n", saddr2str(p->local));
+		plog(LLV_DEBUG2, LOCATION, NULL, "p->remote: %s\n", saddr2str(p->remote));
 		if (CMPSADDR(local, p->local) == 0
-		 && CMPSADDR(remote, p->remote) == 0)
+			&& CMPSADDR(remote, p->remote) == 0){
+			plog(LLV_DEBUG2, LOCATION, NULL, "matched\n");
 			return p;
+		}
 	}
 
+	plog(LLV_DEBUG2, LOCATION, NULL, "no match\n");
+
 	return NULL;
 }
 
@@ -259,9 +278,11 @@ void
 delph1(iph1)
 	struct ph1handle *iph1;
 {
+	if (iph1 == NULL)
+		return;
+
 	/* SA down shell script hook */
-	if (iph1 != NULL)
-		script_hook(iph1, SCRIPT_PHASE1_DOWN);
+	script_hook(iph1, SCRIPT_PHASE1_DOWN);
 
 	EVT_PUSH(iph1->local, iph1->remote, EVTT_PHASE1_DOWN, NULL);
 
@@ -275,9 +296,13 @@ delph1(iph1)
 	}
 #endif
 
+#ifdef ENABLE_HYBRID
+	if (iph1->mode_cfg)
+		isakmp_cfg_rmstate(iph1);
+#endif
+
 #ifdef ENABLE_DPD
-	if (iph1->dpd_r_u != NULL)
-		SCHED_KILL(iph1->dpd_r_u);
+	SCHED_KILL(iph1->dpd_r_u);
 #endif
 
 	if (iph1->remote) {
@@ -288,17 +313,11 @@ delph1(iph1)
 		racoon_free(iph1->local);
 		iph1->local = NULL;
 	}
-
 	if (iph1->approval) {
 		delisakmpsa(iph1->approval);
 		iph1->approval = NULL;
 	}
 
-#ifdef ENABLE_HYBRID
-	if (iph1->mode_cfg)
-		isakmp_cfg_rmstate(iph1);
-#endif
-
 	VPTRINIT(iph1->authstr);
 
 	sched_scrub_param(iph1);
@@ -332,6 +351,9 @@ delph1(iph1)
 	VPTRINIT(iph1->id);
 	VPTRINIT(iph1->id_p);
 
+	if(iph1->approval != NULL)
+		delisakmpsa(iph1->approval);
+
 	if (iph1->ivm) {
 		oakley_delivm(iph1->ivm);
 		iph1->ivm = NULL;
@@ -468,8 +490,21 @@ getph2byid(src, dst, spid)
 	LIST_FOREACH(p, &ph2tree, chain) {
 		if (spid == p->spid &&
 		    CMPSADDR(src, p->src) == 0 &&
-		    CMPSADDR(dst, p->dst) == 0)
-			return p;
+		    CMPSADDR(dst, p->dst) == 0){
+			/* Sanity check to detect zombie handlers
+			 * XXX Sould be done "somewhere" more interesting,
+			 * because we have lots of getph2byxxxx(), but this one
+			 * is called by pk_recvacquire(), so is the most important.
+			 */
+			if(p->status < PHASE2ST_ESTABLISHED &&
+			   p->retry_counter == 0
+			   && p->sce == NULL && p->scr == NULL){
+				plog(LLV_DEBUG, LOCATION, NULL,
+					 "Zombie ph2 found, expiring it\n");
+				isakmp_ph2expire(p);
+			}else
+				return p;
+		}
 	}
 
 	return NULL;
@@ -667,14 +702,23 @@ flushph2()
 {
 	struct ph2handle *p, *next;
 
+	plog(LLV_DEBUG2, LOCATION, NULL,
+		 "flushing all ph2 handlers...\n");
+
 	for (p = LIST_FIRST(&ph2tree); p; p = next) {
 		next = LIST_NEXT(p, chain);
 
 		/* send delete information */
-		if (p->status == PHASE2ST_ESTABLISHED) 
+		if (p->status == PHASE2ST_ESTABLISHED){
+			plog(LLV_DEBUG2, LOCATION, NULL,
+				 "got a ph2 handler to flush...\n");
 			isakmp_info_send_d2(p);
+		}else{
+			plog(LLV_DEBUG2, LOCATION, NULL,
+				 "skipping ph2 handler (state %d)\n", p->status);
+		}
 
-		delete_spd(p);
+		delete_spd(p, 0);
 		unbindph12(p);
 		remph2(p);
 		delph2(p);
@@ -772,6 +816,12 @@ inscontacted(remote)
 		return -1;
 
 	new->remote = dupsaddr(remote);
+	if (new->remote == NULL) {
+		plog(LLV_ERROR, LOCATION, NULL,
+			"failed to allocate buffer.\n");
+		racoon_free(new);
+		return -1;
+	}
 
 	LIST_INSERT_HEAD(&ctdtree, new, chain);
 
@@ -1011,3 +1061,505 @@ exclude_cfg_addr(addr)
 	return 1;
 }
 #endif
+
+
+
+/* 
+ * Reload conf code
+ */
+static int revalidate_ph2(struct ph2handle *iph2){
+	struct sainfoalg *alg;
+	int found, check_level;
+	struct sainfo *sainfo;
+	struct saprop *approval;
+	struct ph1handle *iph1;
+
+	/* 
+	 * Get the new sainfo using values of the old one
+	 */
+	if (iph2->sainfo != NULL) {
+		iph2->sainfo = getsainfo(iph2->sainfo->idsrc, 
+					  iph2->sainfo->iddst, iph2->sainfo->id_i,
+					  iph2->sainfo->remoteid);
+	}
+	approval = iph2->approval;
+	sainfo = iph2->sainfo;
+
+	if (sainfo == NULL) {
+		/* 
+		 * Sainfo has been removed
+		 */
+		plog(LLV_DEBUG, LOCATION, NULL,
+			 "Reload: No sainfo for ph2\n");
+		return 0;
+	}
+
+	if (approval == NULL) {
+		/*
+		 * XXX why do we have a NULL approval sometimes ???
+		 */
+		plog(LLV_DEBUG, LOCATION, NULL,
+			 "No approval found !\n");
+		return 0;
+	}	
+
+	/*
+	 * Don't care about proposals, should we do something ?
+	 * We have to keep iph2->proposal valid at least for initiator,
+	 * for pk_sendgetspi()
+	 */
+
+	plog(LLV_DEBUG, LOCATION, NULL, "active single bundle:\n");
+	printsaprop0(LLV_DEBUG, approval);
+
+	/*
+	 * Validate approval against sainfo
+	 * Note: we must have an updated ph1->rmconf before doing that,
+	 * we'll set check_level to EXACT if we don't have a ph1
+	 * XXX try tu find the new remote section to get the new check level ?
+	 * XXX lifebyte
+	 */
+	if (iph2->ph1 != NULL)
+		iph1=iph2->ph1;
+	else
+		iph1=getph1byaddr(iph2->src, iph2->dst);
+
+	if(iph1 != NULL && iph1->rmconf != NULL) {
+		check_level = iph1->rmconf->pcheck_level;
+	} else {
+		if(iph1 != NULL)
+			plog(LLV_DEBUG, LOCATION, NULL, "No phase1 rmconf found !\n");
+		else
+			plog(LLV_DEBUG, LOCATION, NULL, "No phase1 found !\n");
+		check_level = PROP_CHECK_EXACT;
+	}
+
+	switch (check_level) {
+	case PROP_CHECK_OBEY:
+		plog(LLV_DEBUG, LOCATION, NULL,
+			 "Reload: OBEY for ph2, ok\n");
+		return 1;
+		break;
+
+	case PROP_CHECK_STRICT:
+		/* FALLTHROUGH */
+	case PROP_CHECK_CLAIM:
+		if (sainfo->lifetime < approval->lifetime) {
+			plog(LLV_DEBUG, LOCATION, NULL,
+				 "Reload: lifetime mismatch\n");
+			return 0;
+		}
+
+#if 0
+		/* Lifebyte is deprecated, just ignore it
+		 */
+		if (sainfo->lifebyte < approval->lifebyte) {
+			plog(LLV_DEBUG, LOCATION, NULL,
+				 "Reload: lifebyte mismatch\n");
+			return 0;
+		}
+#endif
+
+		if (sainfo->pfs_group &&
+		   sainfo->pfs_group != approval->pfs_group) {
+			plog(LLV_DEBUG, LOCATION, NULL,
+				 "Reload: PFS group mismatch\n");
+			return 0;
+		}
+		break;
+
+	case PROP_CHECK_EXACT:
+		if (sainfo->lifetime != approval->lifetime ||
+#if 0
+			/* Lifebyte is deprecated, just ignore it
+			 */
+		    sainfo->lifebyte != approval->lifebyte ||
+#endif
+		    sainfo->pfs_group != iph2->approval->pfs_group) {
+			plog(LLV_DEBUG, LOCATION, NULL,
+			    "Reload: lifetime | pfs mismatch\n");
+			return 0;
+		}
+		break;
+
+	default:
+		plog(LLV_DEBUG, LOCATION, NULL,
+			 "Reload: Shouldn't be here !\n");
+		return 0;
+		break;
+	}
+
+	for (alg = sainfo->algs[algclass_ipsec_auth]; alg; alg = alg->next) {
+		if (alg->alg == approval->head->head->authtype)
+			break;
+	}
+	if (alg == NULL) {
+		plog(LLV_DEBUG, LOCATION, NULL,
+			 "Reload: alg == NULL (auth)\n");
+		return 0;
+	}
+
+	found = 0;
+	for (alg = sainfo->algs[algclass_ipsec_enc]; 
+	    (found == 0 && alg != NULL); alg = alg->next) {
+		plog(LLV_DEBUG, LOCATION, NULL,
+			 "Reload: next ph2 enc alg...\n");
+
+		if (alg->alg != approval->head->head->trns_id){
+			plog(LLV_DEBUG, LOCATION, NULL,
+				 "Reload: encmode mismatch (%d / %d)\n",
+				 alg->alg, approval->head->head->trns_id);
+			continue;
+		}
+
+		switch (check_level){
+		/* PROP_CHECK_STRICT cannot happen here */
+		case PROP_CHECK_EXACT:
+			if (alg->encklen != approval->head->head->encklen) {
+				plog(LLV_DEBUG, LOCATION, NULL,
+					 "Reload: enclen mismatch\n");
+				continue;
+			}
+			break;
+
+		case PROP_CHECK_CLAIM:
+			/* FALLTHROUGH */
+		case PROP_CHECK_STRICT:
+			if (alg->encklen > approval->head->head->encklen) {
+				plog(LLV_DEBUG, LOCATION, NULL,
+					 "Reload: enclen mismatch\n");
+				continue;
+			}
+			break;
+
+		default:
+			plog(LLV_ERROR, LOCATION, NULL, 
+			    "unexpected check_level\n");
+			continue;
+			break;
+		}
+		found = 1;
+	}
+
+	if (!found){
+		plog(LLV_DEBUG, LOCATION, NULL,
+			 "Reload: No valid enc\n");
+		return 0;
+	}
+
+	/*
+	 * XXX comp
+	 */
+	plog(LLV_DEBUG, LOCATION, NULL,
+		 "Reload: ph2 check ok\n");
+
+	return 1;
+}
+
+
+static void 
+remove_ph2(struct ph2handle *iph2)
+{
+	u_int32_t spis[2];
+
+	if(iph2 == NULL)
+		return;
+
+	plog(LLV_DEBUG, LOCATION, NULL,
+		 "Deleting a Ph2...\n");
+
+	if (iph2->status == PHASE2ST_ESTABLISHED)
+		isakmp_info_send_d2(iph2);
+
+	if(iph2->approval != NULL && iph2->approval->head != NULL){
+		spis[0]=iph2->approval->head->spi;
+		spis[1]=iph2->approval->head->spi_p;
+
+		/* purge_ipsec_spi() will do all the work:
+		 * - delete SPIs in kernel
+		 * - delete generated SPD
+		 * - unbind / rem / del ph2
+		 */
+		purge_ipsec_spi(iph2->dst, iph2->approval->head->proto_id,
+						spis, 2);
+	}else{
+		unbindph12(iph2);
+		remph2(iph2);
+		delph2(iph2);
+	}
+}
+
+static void remove_ph1(struct ph1handle *iph1){
+	struct ph2handle *iph2, *iph2_next;
+
+	if(iph1 == NULL)
+		return;
+
+	plog(LLV_DEBUG, LOCATION, NULL,
+		 "Removing PH1...\n");
+
+	if (iph1->status == PHASE1ST_ESTABLISHED){
+		for (iph2 = LIST_FIRST(&iph1->ph2tree); iph2; iph2 = iph2_next) {
+			iph2_next = LIST_NEXT(iph2, chain);
+			remove_ph2(iph2);
+		}
+		isakmp_info_send_d1(iph1);
+	}
+	iph1->status = PHASE1ST_EXPIRED;
+	iph1->sce = sched_new(1, isakmp_ph1delete_stub, iph1);
+}
+
+
+static int revalidate_ph1tree_rmconf(void){
+	struct ph1handle *p, *next;
+	struct remoteconf *newrmconf;
+
+	for (p = LIST_FIRST(&ph1tree); p; p = next) {
+		next = LIST_NEXT(p, chain);
+
+		if (p->status == PHASE1ST_EXPIRED)
+			continue;
+
+		newrmconf=getrmconf(p->remote);
+		if(newrmconf == NULL){
+			p->rmconf = NULL;
+			remove_ph1(p);
+		}else{
+			/* Do not free old rmconf, it is just a pointer to an entry in rmtree
+			 */
+			p->rmconf=newrmconf;
+			if(p->approval != NULL){
+				struct isakmpsa *tmpsa;
+
+				tmpsa=dupisakmpsa(p->approval);
+				if(tmpsa != NULL){
+					delisakmpsa(p->approval);
+					p->approval=tmpsa;
+					p->approval->rmconf=newrmconf;
+				}
+			}
+		}
+	}
+
+	return 1;
+}
+
+
+/* rmconf is already updated here
+ */
+static int revalidate_ph1(struct ph1handle *iph1){
+	struct isakmpsa *p, *approval;
+	struct etypes *e;
+
+	if(iph1 == NULL ||
+	   iph1->approval == NULL ||
+		iph1->rmconf == NULL)
+		return 0;
+
+	approval=iph1->approval;
+
+	for (e = iph1->rmconf->etypes; e != NULL; e = e->next){
+		if (iph1->etype == e->type)
+			break;
+	}
+
+	if (e == NULL){
+		plog(LLV_DEBUG, LOCATION, NULL,
+			 "Reload: Exchange type mismatch\n");
+		return 0;
+	}
+
+	if (iph1->etype == ISAKMP_ETYPE_AGG &&
+	   approval->dh_group != iph1->rmconf->dh_group){
+		plog(LLV_DEBUG, LOCATION, NULL,
+			 "Reload: DH mismatch\n");
+		return 0;
+	}
+
+	for (p=iph1->rmconf->proposal; p != NULL; p=p->next){
+		plog(LLV_DEBUG, LOCATION, NULL,
+			 "Reload: Trying next proposal...\n");
+
+		if(approval->authmethod != p->authmethod){
+			plog(LLV_DEBUG, LOCATION, NULL,
+				 "Reload: Authmethod mismatch\n");
+			continue;
+		}
+
+		if(approval->enctype != p->enctype){
+			plog(LLV_DEBUG, LOCATION, NULL,
+				 "Reload: enctype mismatch\n");
+			continue;
+		}
+
+		switch (iph1->rmconf->pcheck_level) {
+		case PROP_CHECK_OBEY:
+			plog(LLV_DEBUG, LOCATION, NULL,
+				 "Reload: OBEY pcheck level, ok...\n");
+			return 1;
+			break;
+
+		case PROP_CHECK_CLAIM:
+			/* FALLTHROUGH */
+		case PROP_CHECK_STRICT:
+			if (approval->encklen < p->encklen) {
+				plog(LLV_DEBUG, LOCATION, NULL,
+					 "Reload: encklen mismatch\n");
+				continue;
+			}
+
+			if (approval->lifetime > p->lifetime) {
+				plog(LLV_DEBUG, LOCATION, NULL,
+					 "Reload: lifetime mismatch\n");
+				continue;
+			}
+
+#if 0
+			/* Lifebyte is deprecated, just ignore it
+			 */
+			if (approval->lifebyte > p->lifebyte) {
+				plog(LLV_DEBUG, LOCATION, NULL,
+					 "Reload: lifebyte mismatch\n");
+				continue;
+			}
+#endif
+			break;
+
+		case PROP_CHECK_EXACT:
+			if (approval->encklen != p->encklen) {
+				plog(LLV_DEBUG, LOCATION, NULL,
+					 "Reload: encklen mismatch\n");
+				continue;
+			}
+
+			if (approval->lifetime != p->lifetime) {
+				plog(LLV_DEBUG, LOCATION, NULL,
+					 "Reload: lifetime mismatch\n");
+				continue;
+			}
+
+#if 0
+			/* Lifebyte is deprecated, just ignore it
+			 */
+			if (approval->lifebyte != p->lifebyte) {
+				plog(LLV_DEBUG, LOCATION, NULL,
+					 "Reload: lifebyte mismatch\n");
+				continue;
+			}
+#endif
+			break;
+
+		default:
+			plog(LLV_ERROR, LOCATION, NULL, 
+			    "unexpected check_level\n");
+			continue;
+			break;
+		}
+
+		if (approval->hashtype != p->hashtype) {
+			plog(LLV_DEBUG, LOCATION, NULL,
+				 "Reload: hashtype mismatch\n");
+			continue;
+		}
+
+		if (iph1->etype != ISAKMP_ETYPE_AGG &&
+		    approval->dh_group != p->dh_group) {
+			plog(LLV_DEBUG, LOCATION, NULL,
+				 "Reload: dhgroup mismatch\n");
+			continue;
+		}
+
+		plog(LLV_DEBUG, LOCATION, NULL, "Reload: Conf ok\n");
+		return 1;
+	}
+
+	plog(LLV_DEBUG, LOCATION, NULL, "Reload: No valid conf found\n");
+	return 0;
+}
+
+
+static int revalidate_ph1tree(void){
+	struct ph1handle *p, *next;
+
+	for (p = LIST_FIRST(&ph1tree); p; p = next) {
+		next = LIST_NEXT(p, chain);
+
+		if (p->status == PHASE1ST_EXPIRED)
+			continue;
+
+		if(!revalidate_ph1(p))
+			remove_ph1(p);
+	}
+
+	return 1;
+}
+
+static int revalidate_ph2tree(void){
+	struct ph2handle *p, *next;
+
+	for (p = LIST_FIRST(&ph2tree); p; p = next) {
+		next = LIST_NEXT(p, chain);
+
+		if (p->status == PHASE2ST_EXPIRED)
+			continue;
+
+		if(!revalidate_ph2(p)){
+			plog(LLV_DEBUG, LOCATION, NULL,
+				 "PH2 not validated, removing it\n");
+			remove_ph2(p);
+		}
+	}
+
+	return 1;
+}
+
+int 
+revalidate_ph12(void)
+{
+
+	revalidate_ph1tree_rmconf();
+
+	revalidate_ph2tree();
+	revalidate_ph1tree();
+
+	return 1;
+}
+
+#ifdef ENABLE_HYBRID
+struct ph1handle *
+getph1bylogin(login)
+	char *login;
+{
+	struct ph1handle *p;
+
+	LIST_FOREACH(p, &ph1tree, chain) {
+		if (p->mode_cfg == NULL)
+			continue;
+		if (strncmp(p->mode_cfg->login, login, LOGINLEN) == 0)
+			return p;
+	}
+
+	return NULL;
+}
+
+int
+purgeph1bylogin(login)
+	char *login;
+{
+	struct ph1handle *p;
+	int found = 0;
+
+	LIST_FOREACH(p, &ph1tree, chain) {
+		if (p->mode_cfg == NULL)
+			continue;
+		if (strncmp(p->mode_cfg->login, login, LOGINLEN) == 0) {
+			if (p->status == PHASE1ST_ESTABLISHED)
+				isakmp_info_send_d1(p);
+			purge_remote(p);
+			found++;
+		}
+	}
+
+	return found;
+}
+#endif
