@@ -1,4 +1,6 @@
-/* $Id: isakmp_inf.c,v 1.14.4.9 2005/08/02 15:09:26 vanhu Exp $ */
+/*	$NetBSD: isakmp_inf.c,v 1.14.4.8 2007/08/01 11:52:20 vanhu Exp $	*/
+
+/* Id: isakmp_inf.c,v 1.44 2006/05/06 20:45:52 manubsd Exp */
 
 /*
  * Copyright (C) 1995, 1996, 1997, and 1998 WIDE Project.
@@ -38,11 +40,7 @@
 #include <net/pfkeyv2.h>
 #include <netinet/in.h>
 #include <sys/queue.h>
-#ifndef HAVE_NETINET6_IPSEC
-#include <netinet/ipsec.h>
-#else
-#include <netinet6/ipsec.h>
-#endif
+#include PATH_IPSEC_H
 
 #include <stdlib.h>
 #include <stdio.h>
@@ -58,6 +56,9 @@
 #  include <time.h>
 # endif
 #endif
+#ifdef ENABLE_HYBRID
+#include <resolv.h>
+#endif
 
 #include "libpfkey.h"
 
@@ -72,16 +73,19 @@
 #include "localconf.h"
 #include "remoteconf.h"
 #include "sockmisc.h"
+#include "handler.h"
+#include "policy.h"
+#include "proposal.h"
 #include "isakmp_var.h"
 #include "evt.h"
 #include "isakmp.h"
 #ifdef ENABLE_HYBRID
 #include "isakmp_xauth.h"
+#include "isakmp_unity.h"
 #include "isakmp_cfg.h" 
 #endif
 #include "isakmp_inf.h"
 #include "oakley.h"
-#include "handler.h"
 #include "ipsec_doi.h"
 #include "crypto_openssl.h"
 #include "pfkey.h"
@@ -95,8 +99,8 @@
 #endif
 
 /* information exchange */
-static int isakmp_info_recv_n __P((struct ph1handle *, vchar_t *));
-static int isakmp_info_recv_d __P((struct ph1handle *, vchar_t *));
+static int isakmp_info_recv_n (struct ph1handle *, struct isakmp_pl_n *, u_int32_t, int);
+static int isakmp_info_recv_d (struct ph1handle *, struct isakmp_pl_d *, u_int32_t, int);
 
 #ifdef ENABLE_DPD
 static int isakmp_info_recv_r_u __P((struct ph1handle *,
@@ -107,7 +111,6 @@ static void isakmp_info_send_r_u __P((void *));
 #endif
 
 static void purge_isakmp_spi __P((int, isakmp_index *, size_t));
-static void purge_ipsec_spi __P((struct sockaddr *, int, u_int32_t *, size_t));
 static void info_recv_initialcontact __P((struct ph1handle *));
 
 /* %%%
@@ -122,24 +125,33 @@ isakmp_info_recv(iph1, msg0)
 	vchar_t *msg0;
 {
 	vchar_t *msg = NULL;
+	vchar_t *pbuf = NULL;
+	u_int32_t msgid = 0;
 	int error = -1;
 	struct isakmp *isakmp;
 	struct isakmp_gen *gen;
+	struct isakmp_parse_t *pa, *pap;
 	void *p;
 	vchar_t *hash, *payload;
 	struct isakmp_gen *nd;
 	u_int8_t np;
 	int encrypted;
+	int flag;
 
 	plog(LLV_DEBUG, LOCATION, NULL, "receive Information.\n");
 
 	encrypted = ISSET(((struct isakmp *)msg0->v)->flags, ISAKMP_FLAG_E);
+	msgid = ((struct isakmp *)msg0->v)->msgid;
 
 	/* Use new IV to decrypt Informational message. */
 	if (encrypted) {
-
 		struct isakmp_ivm *ivm;
 
+		if (iph1->ivm == NULL) {
+			plog(LLV_ERROR, LOCATION, NULL, "iph1->ivm == NULL\n");
+			return -1;
+		}
+
 		/* compute IV */
 		ivm = oakley_newiv2(iph1, ((struct isakmp *)msg0->v)->msgid);
 		if (ivm == NULL)
@@ -168,18 +180,18 @@ isakmp_info_recv(iph1, msg0)
 	if (encrypted) {
 		if (isakmp->np != ISAKMP_NPTYPE_HASH) {
 			plog(LLV_ERROR, LOCATION, NULL,
-			    "ignore information because the "
+			    "ignore information because the"
 			    "message has no hash payload.\n");
 			goto end;
 		}
 
 		if (iph1->status != PHASE1ST_ESTABLISHED) {
 			plog(LLV_ERROR, LOCATION, NULL,
-			    "ignore information because ISAKMP-SA "
+			    "ignore information because ISAKMP-SA"
 			    "has not been established yet.\n");
 			goto end;
 		}
-
+		
 		/* Safety check */
 		if (msg->l < sizeof(*isakmp) + ntohs(gen->len) + sizeof(*nd)) {
 			plog(LLV_ERROR, LOCATION, NULL, 
@@ -247,7 +259,7 @@ isakmp_info_recv(iph1, msg0)
 		vfree(hash);
 		vfree(payload);
 	} else {
-		/* make sure the packet were encrypted after the beginning of phase 1. */
+		/* make sure the packet was encrypted after the beginning of phase 1. */
 		switch (iph1->etype) {
 		case ISAKMP_ETYPE_AGG:
 		case ISAKMP_ETYPE_BASE:
@@ -261,38 +273,304 @@ isakmp_info_recv(iph1, msg0)
 			plog(LLV_ERROR, LOCATION, iph1->remote,
 				"%s message must be encrypted\n",
 				s_isakmp_nptype(np));
+			error = 0;
 			goto end;
 		}
 	}
 
-	switch (np) {
-	case ISAKMP_NPTYPE_N:
-		if (isakmp_info_recv_n(iph1, msg) < 0)
-			goto end;
-		break;
-	case ISAKMP_NPTYPE_D:
-		if (isakmp_info_recv_d(iph1, msg) < 0)
-			goto end;
-		break;
-	case ISAKMP_NPTYPE_NONCE:
-		/* XXX to be 6.4.2 ike-01.txt */
-		/* XXX IV is to be synchronized. */
-		plog(LLV_ERROR, LOCATION, iph1->remote,
-			"ignore Acknowledged Informational\n");
-		break;
-	default:
-		/* don't send information, see isakmp_ident_r1() */
-		error = 0;
-		plog(LLV_ERROR, LOCATION, iph1->remote,
-			"reject the packet, "
-			"received unexpecting payload type %d.\n",
-			gen->np);
+	if (!(pbuf = isakmp_parse(msg))) {
+		error = -1;
 		goto end;
 	}
 
+	error = 0;
+	for (pa = (struct isakmp_parse_t *)pbuf->v; pa->type; pa++) {
+		switch (pa->type) {
+		case ISAKMP_NPTYPE_HASH:
+			/* Handled above */
+			break;
+		case ISAKMP_NPTYPE_N:
+			error = isakmp_info_recv_n(iph1,
+				(struct isakmp_pl_n *)pa->ptr,
+				msgid, encrypted);
+			break;
+		case ISAKMP_NPTYPE_D:
+			error = isakmp_info_recv_d(iph1,
+				(struct isakmp_pl_d *)pa->ptr,
+				msgid, encrypted);
+			break;
+		case ISAKMP_NPTYPE_NONCE:
+			/* XXX to be 6.4.2 ike-01.txt */
+			/* XXX IV is to be synchronized. */
+			plog(LLV_ERROR, LOCATION, iph1->remote,
+				"ignore Acknowledged Informational\n");
+			break;
+		default:
+			/* don't send information, see isakmp_ident_r1() */
+			error = 0;
+			plog(LLV_ERROR, LOCATION, iph1->remote,
+				"reject the packet, "
+				"received unexpected payload type %s.\n",
+				s_isakmp_nptype(gen->np));
+		}
+		if(error < 0) {
+			break;
+		} else {
+			flag |= error;
+		}
+	}
     end:
 	if (msg != NULL)
 		vfree(msg);
+	if (pbuf != NULL)
+		vfree(pbuf);
+	return error;
+}
+
+/*
+ * handling of Notification payload
+ */
+static int
+isakmp_info_recv_n(iph1, notify, msgid, encrypted)
+	struct ph1handle *iph1;
+	struct isakmp_pl_n *notify;
+	u_int32_t msgid;
+	int encrypted;
+{
+	u_int type;
+	vchar_t *pbuf;
+	vchar_t *ndata;
+	char *nraw;
+	size_t l;
+	char *spi;
+
+	type = ntohs(notify->type);
+
+	switch (type) {
+	case ISAKMP_NTYPE_CONNECTED:
+	case ISAKMP_NTYPE_RESPONDER_LIFETIME:
+	case ISAKMP_NTYPE_REPLAY_STATUS:
+#ifdef ENABLE_HYBRID
+	case ISAKMP_NTYPE_UNITY_HEARTBEAT:
+#endif
+		/* do something */
+		break;
+	case ISAKMP_NTYPE_INITIAL_CONTACT:
+		if (encrypted)
+			info_recv_initialcontact(iph1);
+			return 0;
+		break;
+#ifdef ENABLE_DPD
+	case ISAKMP_NTYPE_R_U_THERE:
+		if (encrypted)
+			return isakmp_info_recv_r_u(iph1,
+				(struct isakmp_pl_ru *)notify, msgid);
+		break;
+	case ISAKMP_NTYPE_R_U_THERE_ACK:
+		if (encrypted)
+			return isakmp_info_recv_r_u_ack(iph1,
+				(struct isakmp_pl_ru *)notify, msgid);
+		break;
+#endif
+	default:
+	    {
+		/* XXX there is a potential of dos attack. */
+		if(type >= ISAKMP_NTYPE_MINERROR &&
+		   type <= ISAKMP_NTYPE_MAXERROR) {
+			if (msgid == 0) {
+				/* don't think this realy deletes ph1 ? */
+				plog(LLV_ERROR, LOCATION, iph1->remote,
+					"delete phase1 handle.\n");
+				return -1;
+			} else {
+				if (getph2bymsgid(iph1, msgid) == NULL) {
+					plog(LLV_ERROR, LOCATION, iph1->remote,
+						"fatal %s notify messsage, "
+						"phase1 should be deleted.\n",
+						s_isakmp_notify_msg(type));
+				} else {
+					plog(LLV_ERROR, LOCATION, iph1->remote,
+						"fatal %s notify messsage, "
+						"phase2 should be deleted.\n",
+						s_isakmp_notify_msg(type));
+				}
+			}
+		} else {
+			plog(LLV_ERROR, LOCATION, iph1->remote,
+				"unhandled notify message %s, "
+				"no phase2 handle found.\n",
+				s_isakmp_notify_msg(type));
+		}
+	    }
+	    break;
+	}
+
+	/* get spi if specified and allocate */
+	if(notify->spi_size > 0) {
+		if (ntohs(notify->h.len) < sizeof(*notify) + notify->spi_size) {
+			plog(LLV_ERROR, LOCATION, iph1->remote,
+				"invalid spi_size in notification payload.\n");
+			return -1;
+		}
+		spi = val2str((char *)(notify + 1), notify->spi_size);
+
+		plog(LLV_DEBUG, LOCATION, iph1->remote,
+			"notification message %d:%s, "
+			"doi=%d proto_id=%d spi=%s(size=%d).\n",
+			type, s_isakmp_notify_msg(type),
+			ntohl(notify->doi), notify->proto_id, spi, notify->spi_size);
+
+		racoon_free(spi);
+	}
+
+	/* Send the message data to the logs */
+	if(type >= ISAKMP_NTYPE_MINERROR &&
+	   type <= ISAKMP_NTYPE_MAXERROR) {
+		l = ntohs(notify->h.len) - sizeof(*notify) - notify->spi_size;
+		if (l > 0) {
+			nraw = (char*)notify;	
+			nraw += sizeof(*notify) + notify->spi_size;
+			if ((ndata = vmalloc(l)) != NULL) {
+				memcpy(ndata->v, nraw, ndata->l);
+				plog(LLV_ERROR, LOCATION, iph1->remote,
+				    "Message: '%s'.\n", 
+				    binsanitize(ndata->v, ndata->l));
+				vfree(ndata);
+			} else {
+				plog(LLV_ERROR, LOCATION, iph1->remote,
+				    "Cannot allocate memory\n");
+			}
+		}
+	}
+	return 0;
+}
+
+/*
+ * handling of Deletion payload
+ */
+static int
+isakmp_info_recv_d(iph1, delete, msgid, encrypted)
+	struct ph1handle *iph1;
+	struct isakmp_pl_d *delete;
+	u_int32_t msgid;
+	int encrypted;
+{
+	int tlen, num_spi;
+	vchar_t *pbuf;
+	int protected = 0;
+	struct ph1handle *del_ph1;
+	struct ph2handle *iph2;
+	union {
+		u_int32_t spi32;
+		u_int16_t spi16[2];
+	} spi;
+
+	if (ntohl(delete->doi) != IPSEC_DOI) {
+		plog(LLV_ERROR, LOCATION, iph1->remote,
+			"delete payload with invalid doi:%d.\n",
+			ntohl(delete->doi));
+#ifdef ENABLE_HYBRID
+		/*
+		 * At deconnexion time, Cisco VPN client does this
+		 * with a zero DOI. Don't give up in that situation.
+		 */
+		if (((iph1->mode_cfg->flags &
+		    ISAKMP_CFG_VENDORID_UNITY) == 0) || (delete->doi != 0))
+			return 0;
+#else
+		return 0;
+#endif
+	}
+
+	num_spi = ntohs(delete->num_spi);
+	tlen = ntohs(delete->h.len) - sizeof(struct isakmp_pl_d);
+
+	if (tlen != num_spi * delete->spi_size) {
+		plog(LLV_ERROR, LOCATION, iph1->remote,
+			"deletion payload with invalid length.\n");
+		return 0;
+	}
+
+	plog(LLV_DEBUG, LOCATION, iph1->remote,
+		"delete payload for protocol %s\n",
+		s_ipsecdoi_proto(delete->proto_id));
+
+	if(!iph1->rmconf->weak_phase1_check && !encrypted) {
+		plog(LLV_WARNING, LOCATION, iph1->remote,
+			"Ignoring unencrypted delete payload "
+			"(check the weak_phase1_check option)\n");
+		return 0;
+	}
+
+	switch (delete->proto_id) {
+	case IPSECDOI_PROTO_ISAKMP:
+		if (delete->spi_size != sizeof(isakmp_index)) {
+			plog(LLV_ERROR, LOCATION, iph1->remote,
+				"delete payload with strange spi "
+				"size %d(proto_id:%d)\n",
+				delete->spi_size, delete->proto_id);
+			return 0;
+		}
+
+		del_ph1=getph1byindex((isakmp_index *)(delete + 1));
+		if(del_ph1 != NULL){
+
+			EVT_PUSH(del_ph1->local, del_ph1->remote,
+			EVTT_PEERPH1_NOPROP, NULL);
+			if (del_ph1->scr)
+				SCHED_KILL(del_ph1->scr);
+
+			/*
+			 * Do not delete IPsec SAs when receiving an IKE delete notification.
+			 * Just delete the IKE SA.
+			 */
+			isakmp_ph1expire(del_ph1);
+		}
+		break;
+
+	case IPSECDOI_PROTO_IPSEC_AH:
+	case IPSECDOI_PROTO_IPSEC_ESP:
+		if (delete->spi_size != sizeof(u_int32_t)) {
+			plog(LLV_ERROR, LOCATION, iph1->remote,
+				"delete payload with strange spi "
+				"size %d(proto_id:%d)\n",
+				delete->spi_size, delete->proto_id);
+			return 0;
+		}
+		EVT_PUSH(iph1->local, iph1->remote, 
+		    EVTT_PEER_DELETE, NULL);
+		purge_ipsec_spi(iph1->remote, delete->proto_id,
+		    (u_int32_t *)(delete + 1), num_spi);
+		break;
+
+	case IPSECDOI_PROTO_IPCOMP:
+		/* need to handle both 16bit/32bit SPI */
+		memset(&spi, 0, sizeof(spi));
+		if (delete->spi_size == sizeof(spi.spi16[1])) {
+			memcpy(&spi.spi16[1], delete + 1,
+			    sizeof(spi.spi16[1]));
+		} else if (delete->spi_size == sizeof(spi.spi32))
+			memcpy(&spi.spi32, delete + 1, sizeof(spi.spi32));
+		else {
+			plog(LLV_ERROR, LOCATION, iph1->remote,
+				"delete payload with strange spi "
+				"size %d(proto_id:%d)\n",
+				delete->spi_size, delete->proto_id);
+			return 0;
+		}
+		purge_ipsec_spi(iph1->remote, delete->proto_id,
+		    &spi.spi32, num_spi);
+		break;
+
+	default:
+		plog(LLV_ERROR, LOCATION, iph1->remote,
+			"deletion message received, "
+			"invalid proto_id: %d\n",
+			delete->proto_id);
+		return 0;
+	}
+
+	plog(LLV_DEBUG, LOCATION, NULL, "purged SAs.\n");
 
 	return 0;
 }
@@ -364,8 +642,11 @@ isakmp_info_send_d2(iph2)
 	 * It's nonsensical to negotiate phase 1 to send the information.
 	 */
 	iph1 = getph1byaddr(iph2->src, iph2->dst); 
-	if (iph1 == NULL)
+	if (iph1 == NULL){
+		plog(LLV_DEBUG2, LOCATION, NULL,
+			 "No ph1 handler found, could not send DELETE_SA\n");
 		return 0;
+	}
 
 	/* create delete payload */
 	for (pr = iph2->approval->head; pr != NULL; pr = pr->next) {
@@ -449,8 +730,10 @@ isakmp_info_send_nx(isakmp, remote, local, type, data)
 	iph1->flags = 0;
 	iph1->msgid = 0;	/* XXX */
 #ifdef ENABLE_HYBRID
-	if ((iph1->mode_cfg = isakmp_cfg_mkstate()) == NULL)
-		return -1;
+	if ((iph1->mode_cfg = isakmp_cfg_mkstate()) == NULL) {
+		error = -1;
+		goto end;
+	}
 #endif
 #ifdef ENABLE_FRAG
 	iph1->frag = 0;
@@ -458,8 +741,10 @@ isakmp_info_send_nx(isakmp, remote, local, type, data)
 #endif
 
 	/* copy remote address */
-	if (copy_ph1addresses(iph1, rmconf, remote, local) < 0)
-		return -1;
+	if (copy_ph1addresses(iph1, rmconf, remote, local) < 0) {
+		error = -1;
+		goto end;
+	}
 
 	tlen = sizeof(*n) + spisiz;
 	if (data)
@@ -468,6 +753,7 @@ isakmp_info_send_nx(isakmp, remote, local, type, data)
 	if (payload == NULL) { 
 		plog(LLV_ERROR, LOCATION, NULL,
 			"failed to get buffer to send.\n");
+		error = -1;
 		goto end;
 	}
 
@@ -479,7 +765,7 @@ isakmp_info_send_nx(isakmp, remote, local, type, data)
 	n->spi_size = spisiz;
 	n->type = htons(type);
 	if (spisiz)
-		memset(n + 1, 0, spisiz);	/*XXX*/
+		memset(n + 1, 0, spisiz);	/* XXX spisiz is always 0 */
 	if (data)
 		memcpy((caddr_t)(n + 1) + spisiz, data->v, data->l);
 
@@ -626,18 +912,28 @@ isakmp_info_send_common(iph1, payload, np, flags)
 		goto end;
 
 	iph2->dst = dupsaddr(iph1->remote);
+	if (iph2->dst == NULL) {
+		delph2(iph2);
+		goto end;
+	}
 	iph2->src = dupsaddr(iph1->local);
+	if (iph2->src == NULL) {
+		delph2(iph2);
+		goto end;
+	}
 	switch (iph1->remote->sa_family) {
 	case AF_INET:
-#ifndef ENABLE_NATT
+#if (!defined(ENABLE_NATT)) || (defined(BROKEN_NATT))
 		((struct sockaddr_in *)iph2->dst)->sin_port = 0;
 		((struct sockaddr_in *)iph2->src)->sin_port = 0;
 #endif
 		break;
 #ifdef INET6
 	case AF_INET6:
+#if (!defined(ENABLE_NATT)) || (defined(BROKEN_NATT))
 		((struct sockaddr_in6 *)iph2->dst)->sin6_port = 0;
 		((struct sockaddr_in6 *)iph2->src)->sin6_port = 0;
+#endif
 		break;
 #endif
 	default:
@@ -821,113 +1117,7 @@ isakmp_add_pl_n(buf0, np_p, type, pr, data)
 	return buf;
 }
 
-/*
- * handling to receive Notification payload
- */
-static int
-isakmp_info_recv_n(iph1, msg)
-	struct ph1handle *iph1;
-	vchar_t *msg;
-{
-	struct isakmp_pl_n *n = NULL;
-	u_int type;
-	vchar_t *pbuf;
-	struct isakmp_parse_t *pa, *pap;
-	char *spi;
-
-	if (!(pbuf = isakmp_parse(msg)))
-		return -1;
-	pa = (struct isakmp_parse_t *)pbuf->v;
-	for (pap = pa; pap->type; pap++) {
-		switch (pap->type) {
-		case ISAKMP_NPTYPE_HASH:
-			/* do something here */
-			break;
-		case ISAKMP_NPTYPE_NONCE:
-			/* send to ack */
-			break;
-		case ISAKMP_NPTYPE_N:
-			n = (struct isakmp_pl_n *)pap->ptr;
-			break;
-		default:
-			vfree(pbuf);
-			return -1;
-		}
-	}
-	vfree(pbuf);
-	if (!n)
-		return -1;
-
-	type = ntohs(n->type);
-
-	switch (type) {
-	case ISAKMP_NTYPE_CONNECTED:
-	case ISAKMP_NTYPE_RESPONDER_LIFETIME:
-	case ISAKMP_NTYPE_REPLAY_STATUS:
-		/* do something */
-		break;
-	case ISAKMP_NTYPE_INITIAL_CONTACT:
-		info_recv_initialcontact(iph1);
-		break;
-#ifdef ENABLE_DPD
-	case ISAKMP_NTYPE_R_U_THERE:
-		isakmp_info_recv_r_u(iph1, (struct isakmp_pl_ru *)n,
-				     ((struct isakmp *)msg->v)->msgid);
-		break;
-	case ISAKMP_NTYPE_R_U_THERE_ACK:
-		isakmp_info_recv_r_u_ack(iph1, (struct isakmp_pl_ru *)n,
-					 ((struct isakmp *)msg->v)->msgid);
-		break;
-#endif
-
-	default:
-	    {
-		u_int32_t msgid = ((struct isakmp *)msg->v)->msgid;
-		struct ph2handle *iph2;
-
-		/* XXX there is a potential of dos attack. */
-		if (msgid == 0) {
-			/* delete ph1 */
-			plog(LLV_ERROR, LOCATION, iph1->remote,
-				"delete phase1 handle.\n");
-			return -1;
-		} else {
-			iph2 = getph2bymsgid(iph1, msgid);
-			if (iph2 == NULL) {
-				plog(LLV_ERROR, LOCATION, iph1->remote,
-					"unknown notify message, "
-					"no phase2 handle found.\n");
-			} else {
-				/* delete ph2 */
-				unbindph12(iph2);
-				remph2(iph2);
-				delph2(iph2);
-			}
-		}
-	    }
-		break;
-	}
-
-	/* get spi and allocate */
-	if (ntohs(n->h.len) < sizeof(*n) + n->spi_size) {
-		plog(LLV_ERROR, LOCATION, iph1->remote,
-			"invalid spi_size in notification payload.\n");
-		return -1;
-	}
-	spi = val2str((char *)(n + 1), n->spi_size);
-
-	plog(LLV_DEBUG, LOCATION, iph1->remote,
-		"notification message %d:%s, "
-		"doi=%d proto_id=%d spi=%s(size=%d).\n",
-		type, s_isakmp_notify_msg(type),
-		ntohl(n->doi), n->proto_id, spi, n->spi_size);
-
-	racoon_free(spi);
-
-	return(0);
-}
-
-void
+static void
 purge_isakmp_spi(proto, spi, n)
 	int proto;
 	isakmp_index *spi;	/*network byteorder*/
@@ -946,14 +1136,15 @@ purge_isakmp_spi(proto, spi, n)
 			s_ipsecdoi_proto(proto),
 			isakmp_pindex(&spi[i], 0));
 
-		if (iph1->sce)
-			SCHED_KILL(iph1->sce);
+		SCHED_KILL(iph1->sce);
 		iph1->status = PHASE1ST_EXPIRED;
 		iph1->sce = sched_new(1, isakmp_ph1delete_stub, iph1);
 	}
 }
 
-static void
+
+
+void
 purge_ipsec_spi(dst0, proto, spi, n)
 	struct sockaddr *dst0;
 	int proto;
@@ -963,11 +1154,18 @@ purge_ipsec_spi(dst0, proto, spi, n)
 	vchar_t *buf = NULL;
 	struct sadb_msg *msg, *next, *end;
 	struct sadb_sa *sa;
+	struct sadb_lifetime *lt;
 	struct sockaddr *src, *dst;
 	struct ph2handle *iph2;
+	u_int64_t created;
 	size_t i;
 	caddr_t mhp[SADB_EXT_MAX + 1];
 
+	plog(LLV_DEBUG2, LOCATION, NULL,
+		 "purge_ipsec_spi:\n");
+	plog(LLV_DEBUG2, LOCATION, NULL, "dst0: %s\n", saddr2str(dst0));
+	plog(LLV_DEBUG2, LOCATION, NULL, "SPI: %08X\n", ntohl(spi[0]));
+
 	buf = pfkey_dump_sadb(ipsecdoi2pfkey_proto(proto));
 	if (buf == NULL) {
 		plog(LLV_DEBUG, LOCATION, NULL,
@@ -1003,12 +1201,21 @@ purge_ipsec_spi(dst0, proto, spi, n)
 		}
 		src = PFKEY_ADDR_SADDR(mhp[SADB_EXT_ADDRESS_SRC]);
 		dst = PFKEY_ADDR_SADDR(mhp[SADB_EXT_ADDRESS_DST]);
+		lt = (struct sadb_lifetime*)mhp[SADB_EXT_LIFETIME_HARD];
+		if(lt != NULL)
+			created = lt->sadb_lifetime_addtime;
+		else
+			created = 0;
 
 		if (sa->sadb_sa_state != SADB_SASTATE_MATURE
 		 && sa->sadb_sa_state != SADB_SASTATE_DYING) {
 			msg = next;
 			continue;
 		}
+		plog(LLV_DEBUG2, LOCATION, NULL, "src: %s\n", saddr2str(src));
+		plog(LLV_DEBUG2, LOCATION, NULL, "dst: %s\n", saddr2str(dst));
+
+
 
 		/* XXX n^2 algorithm, inefficient */
 
@@ -1037,8 +1244,8 @@ purge_ipsec_spi(dst0, proto, spi, n)
 			 * exists.
 			 */
 			iph2 = getph2bysaidx(src, dst, proto, spi[i]);
-			if (iph2) {
-				delete_spd(iph2);
+			if(iph2 != NULL){
+				delete_spd(iph2, created);
 				unbindph12(iph2);
 				remph2(iph2);
 				delph2(iph2);
@@ -1083,8 +1290,10 @@ info_recv_initialcontact(iph1)
 		return;
 
 #if 0
-	loc = strdup(saddrwop2str(iph1->local));
-	rem = strdup(saddrwop2str(iph1->remote));
+	loc = racoon_strdup(saddrwop2str(iph1->local));
+	rem = racoon_strdup(saddrwop2str(iph1->remote));
+	STRDUP_FATAL(loc);
+	STRDUP_FATAL(rem);
 
 	/*
 	 * Purge all IPSEC-SAs for the peer.  We can do this
@@ -1247,7 +1456,7 @@ info_recv_initialcontact(iph1)
 		proto_id = pfkey2ipsecdoi_proto(msg->sadb_msg_satype);
 		iph2 = getph2bysaidx(src, dst, proto_id, sa->sadb_sa_spi);
 		if (iph2) {
-			delete_spd(iph2);
+			delete_spd(iph2, 0);
 			unbindph12(iph2);
 			remph2(iph2);
 			delph2(iph2);
@@ -1259,162 +1468,6 @@ info_recv_initialcontact(iph1)
 	vfree(buf);
 }
 
-/*
- * handling to receive Deletion payload
- */
-static int
-isakmp_info_recv_d(iph1, msg)
-	struct ph1handle *iph1;
-	vchar_t *msg;
-{
-	struct isakmp_pl_d *d;
-	int tlen, num_spi;
-	vchar_t *pbuf;
-	struct isakmp_parse_t *pa, *pap;
-	int protected = 0;
-	union {
-		u_int32_t spi32;
-		u_int16_t spi16[2];
-	} spi;
-
-	/* validate the type of next payload */
-	if (!(pbuf = isakmp_parse(msg)))
-		return -1;
-	pa = (struct isakmp_parse_t *)pbuf->v;
-	for (pap = pa; pap->type; pap++) {
-		switch (pap->type) {
-		case ISAKMP_NPTYPE_D:
-			break;
-		case ISAKMP_NPTYPE_HASH:
-			if (pap == pa) {
-				protected++;
-				break;
-			}
-			plog(LLV_ERROR, LOCATION, iph1->remote,
-				"received next payload type %d "
-				"in wrong place (must be the first payload).\n",
-				pap->type);
-			vfree(pbuf);
-			return -1;
-		default:
-			/* don't send information, see isakmp_ident_r1() */
-			plog(LLV_ERROR, LOCATION, iph1->remote,
-				"reject the packet, "
-				"received unexpecting payload type %d.\n",
-				pap->type);
-			vfree(pbuf);
-			return 0;
-		}
-	}
-
-	if (!protected) {
-		plog(LLV_ERROR, LOCATION, NULL,
-			"delete payload is not proteted, "
-			"ignored.\n");
-		vfree(pbuf);
-		return -1;
-	}
-
-	/* process a delete payload */
-	for (pap = pa; pap->type; pap++) {
-		if (pap->type != ISAKMP_NPTYPE_D)
-			continue;
-
-		d = (struct isakmp_pl_d *)pap->ptr;
-
-		if (ntohl(d->doi) != IPSEC_DOI) {
-			plog(LLV_ERROR, LOCATION, iph1->remote,
-				"delete payload with invalid doi:%d.\n",
-				ntohl(d->doi));
-#ifdef ENABLE_HYBRID
-			/*
-			 * At deconnexion time, Cisco VPN client does this
-			 * with a zero DOI. Don't give up in that situation.
-			 */
-			if (((iph1->mode_cfg->flags &
-			    ISAKMP_CFG_VENDORID_UNITY) == 0) || (d->doi != 0))
-				continue;
-#else
-			continue;
-#endif
-		}
-
-		num_spi = ntohs(d->num_spi);
-		tlen = ntohs(d->h.len) - sizeof(struct isakmp_pl_d);
-
-		if (tlen != num_spi * d->spi_size) {
-			plog(LLV_ERROR, LOCATION, iph1->remote,
-				"deletion payload with invalid length.\n");
-			vfree(pbuf);
-			return -1;
-		}
-
-		switch (d->proto_id) {
-		case IPSECDOI_PROTO_ISAKMP:
-			if (d->spi_size != sizeof(isakmp_index)) {
-				plog(LLV_ERROR, LOCATION, iph1->remote,
-					"delete payload with strange spi "
-					"size %d(proto_id:%d)\n",
-					d->spi_size, d->proto_id);
-				continue;
-			}
-
-			if (iph1->scr)
-				SCHED_KILL(iph1->scr);
-
-			purge_remote(iph1);
-			break;
-
-		case IPSECDOI_PROTO_IPSEC_AH:
-		case IPSECDOI_PROTO_IPSEC_ESP:
-			if (d->spi_size != sizeof(u_int32_t)) {
-				plog(LLV_ERROR, LOCATION, iph1->remote,
-					"delete payload with strange spi "
-					"size %d(proto_id:%d)\n",
-					d->spi_size, d->proto_id);
-				continue;
-			}
-			EVT_PUSH(iph1->local, iph1->remote, 
-			    EVTT_PEER_DELETE, NULL);
-			purge_ipsec_spi(iph1->remote, d->proto_id,
-			    (u_int32_t *)(d + 1), num_spi);
-			break;
-
-		case IPSECDOI_PROTO_IPCOMP:
-			/* need to handle both 16bit/32bit SPI */
-			memset(&spi, 0, sizeof(spi));
-			if (d->spi_size == sizeof(spi.spi16[1])) {
-				memcpy(&spi.spi16[1], d + 1,
-				    sizeof(spi.spi16[1]));
-			} else if (d->spi_size == sizeof(spi.spi32))
-				memcpy(&spi.spi32, d + 1, sizeof(spi.spi32));
-			else {
-				plog(LLV_ERROR, LOCATION, iph1->remote,
-					"delete payload with strange spi "
-					"size %d(proto_id:%d)\n",
-					d->spi_size, d->proto_id);
-				continue;
-			}
-			purge_ipsec_spi(iph1->remote, d->proto_id,
-			    &spi.spi32, num_spi);
-			break;
-
-		default:
-			plog(LLV_ERROR, LOCATION, iph1->remote,
-				"deletion message received, "
-				"invalid proto_id: %d\n",
-				d->proto_id);
-			continue;
-		}
-
-		plog(LLV_DEBUG, LOCATION, NULL, "purged SAs.\n");
-	}
-
-	vfree(pbuf);
-
-	return 0;
-}
-
 void
 isakmp_check_notify(gen, iph1)
 	struct isakmp_gen *gen;		/* points to Notify payload */
@@ -1427,31 +1480,26 @@ isakmp_check_notify(gen, iph1)
 
 	switch (ntohs(notify->type)) {
 	case ISAKMP_NTYPE_CONNECTED:
-		plog(LLV_WARNING, LOCATION, iph1->remote,
-			"ignore CONNECTED notification.\n");
-		break;
 	case ISAKMP_NTYPE_RESPONDER_LIFETIME:
-		plog(LLV_WARNING, LOCATION, iph1->remote,
-			"ignore RESPONDER-LIFETIME notification.\n");
-		break;
 	case ISAKMP_NTYPE_REPLAY_STATUS:
+	case ISAKMP_NTYPE_HEARTBEAT:
+#ifdef ENABLE_HYBRID
+	case ISAKMP_NTYPE_UNITY_HEARTBEAT:
+#endif
 		plog(LLV_WARNING, LOCATION, iph1->remote,
-			"ignore REPLAY-STATUS notification.\n");
+			"ignore %s notification.\n",
+			s_isakmp_notify_msg(ntohs(notify->type)));
 		break;
 	case ISAKMP_NTYPE_INITIAL_CONTACT:
 		plog(LLV_WARNING, LOCATION, iph1->remote,
 			"ignore INITIAL-CONTACT notification, "
 			"because it is only accepted after phase1.\n");
 		break;
-	case ISAKMP_NTYPE_HEARTBEAT:
-		plog(LLV_WARNING, LOCATION, iph1->remote,
-			"ignore HEARTBEAT notification\n");
-		break;
 	default:
 		isakmp_info_send_n1(iph1, ISAKMP_NTYPE_INVALID_PAYLOAD_TYPE, NULL);
 		plog(LLV_ERROR, LOCATION, iph1->remote,
-			"received unknown notification type %u.\n",
-		    ntohs(notify->type));
+			"received unknown notification type %s.\n",
+			s_isakmp_notify_msg(ntohs(notify->type)));
 	}
 
 	return;
@@ -1537,8 +1585,7 @@ isakmp_info_recv_r_u_ack (iph1, ru, msgid)
 	/* Useless ??? */
 	iph1->dpd_lastack = time(NULL);
 
-	if (iph1->dpd_r_u != NULL)
-		SCHED_KILL(iph1->dpd_r_u);
+	SCHED_KILL(iph1->dpd_r_u);
 
 	isakmp_sched_r_u(iph1, 0);
 
@@ -1548,8 +1595,10 @@ isakmp_info_recv_r_u_ack (iph1, ru, msgid)
 }
 
 
+
+
 /*
- * send Delete payload (for ISAKMP SA) in Informational exchange.
+ * send DPD R-U-THERE payload in Informational exchange.
  */
 static void
 isakmp_info_send_r_u(arg)
@@ -1579,21 +1628,6 @@ isakmp_info_send_r_u(arg)
 
 	/* TODO: check recent activity to avoid useless sends... */
 
-	/* XXX: why do we have a NULL LIST_FIRST even when a Phase2 exists ??? */
-#if 0
-	if (LIST_FIRST(&iph1->ph2tree) == NULL){
-		/* XXX: No Ph2 => no need to test ph1 ?
-		 */
-		/* Reschedule the r_u_there....
-		   XXX: reschedule when a new ph2 ?
-		 */
-		isakmp_sched_r_u(iph1, 0);
-		plog(LLV_DEBUG, LOCATION, iph1->remote,
-			 "no phase2 handler, rescheduling send_r_u (%d).\n", iph1->rmconf->dpd_interval);
-		return 0;
-	}
-#endif
-
 	tlen = sizeof(*ru);
 	payload = vmalloc(tlen);
 	if (payload == NULL) {
@@ -1657,10 +1691,10 @@ isakmp_sched_r_u(iph1, retry)
 
 	if(retry)
 		iph1->dpd_r_u = sched_new(iph1->rmconf->dpd_retry,
-					  isakmp_info_send_r_u, iph1);
+								  isakmp_info_send_r_u, iph1);
 	else
-		sched_new(iph1->rmconf->dpd_interval,
-			  isakmp_info_send_r_u, iph1);
+		iph1->dpd_r_u = sched_new(iph1->rmconf->dpd_interval,
+								  isakmp_info_send_r_u, iph1);
 
 	return 0;
 }
