@@ -1,4 +1,6 @@
-/* $Id: oakley.c,v 1.17.2.5 2005/10/04 09:54:27 manubsd Exp $ */
+/*	$NetBSD: oakley.c,v 1.9.6.2 2007/04/04 13:08:28 vanhu Exp $	*/
+
+/* Id: oakley.c,v 1.32 2006/05/26 12:19:46 manubsd Exp */
 
 /*
  * Copyright (C) 1995, 1996, 1997, and 1998 WIDE Project.
@@ -36,8 +38,8 @@
 #include <sys/socket.h>	/* XXX for subjectaltname */
 #include <netinet/in.h>	/* XXX for subjectaltname */
 
-#include <openssl/x509.h>
 #include <openssl/pkcs7.h>
+#include <openssl/x509.h>
 
 #include <stdlib.h>
 #include <stdio.h>
@@ -54,6 +56,9 @@
 #  include <time.h>
 # endif
 #endif
+#ifdef ENABLE_HYBRID
+#include <resolv.h>
+#endif
 
 #include "var.h"
 #include "misc.h"
@@ -252,7 +257,7 @@ oakley_dh_compute(dh, pub, priv, pub_p, gxy)
 
 #ifdef ENABLE_STATS
 	gettimeofday(&end, NULL);
-	syslog(LOG_NOTICE, "%s(%s%d): %8.6f", __func__,
+	syslog(LOG_NOTICE, "%s(%s%zu): %8.6f", __func__,
 		s_attr_isakmp_group(dh->type), dh->prime->l << 3,
 		timedelta(&start, &end));
 #endif
@@ -299,7 +304,7 @@ oakley_dh_generate(dh, pub, priv)
 
 #ifdef ENABLE_STATS
 	gettimeofday(&end, NULL);
-	syslog(LOG_NOTICE, "%s(%s%d): %8.6f", __func__,
+	syslog(LOG_NOTICE, "%s(%s%zu): %8.6f", __func__,
 		s_attr_isakmp_group(dh->type), dh->prime->l << 3,
 		timedelta(&start, &end));
 #endif
@@ -588,6 +593,7 @@ oakley_compute_keymat_x(iph2, side, sa_dir)
 
 			while (dupkeymat--) {
 				vchar_t *this = NULL;	/* Kn */
+				int update_prev;
 
 				memcpy(seed->v, prev->v, prev->l);
 				memcpy(seed->v + prev->l, buf->v, buf->l);
@@ -603,8 +609,14 @@ oakley_compute_keymat_x(iph2, side, sa_dir)
 					goto end;
 				}
 
+				update_prev = (prev && prev == res) ? 1 : 0;
+
 				l = res->l;
 				res = vrealloc(res, l + this->l);
+
+				if (update_prev)
+					prev = res;
+
 				if (res == NULL) {
 					plog(LLV_ERROR, LOCATION, NULL,
 						"failed to get keymat buffer.\n");
@@ -846,7 +858,7 @@ oakley_ph1hash_common(iph1, sw)
 		+ (sw == GENERATE ? iph1->id->l : iph1->id_p->l);
 
 #ifdef HAVE_GSSAPI
-	if (iph1->approval->authmethod == OAKLEY_ATTR_AUTH_METHOD_GSSAPI_KRB) {
+	if (AUTHMETHOD(iph1) == OAKLEY_ATTR_AUTH_METHOD_GSSAPI_KRB) {
 		if (iph1->gi_i != NULL && iph1->gi_r != NULL) {
 			bp = (sw == GENERATE ? iph1->gi_i : iph1->gi_r);
 			len += bp->l;
@@ -907,7 +919,7 @@ oakley_ph1hash_common(iph1, sw)
 	p += bp->l;
 
 #ifdef HAVE_GSSAPI
-	if (iph1->approval->authmethod == OAKLEY_ATTR_AUTH_METHOD_GSSAPI_KRB) {
+	if (AUTHMETHOD(iph1) == OAKLEY_ATTR_AUTH_METHOD_GSSAPI_KRB) {
 		if (iph1->gi_i != NULL && iph1->gi_r != NULL) {
 			bp = (sw == GENERATE ? iph1->gi_i : iph1->gi_r);
 			memcpy(p, bp->v, bp->l);
@@ -928,7 +940,8 @@ oakley_ph1hash_common(iph1, sw)
 
 	error = 0;
 
-	plog(LLV_DEBUG, LOCATION, NULL, "HASH computed:\n");
+	plog(LLV_DEBUG, LOCATION, NULL, "HASH (%s) computed:\n",
+		iph1->side == INITIATOR ? "init" : "resp");
 	plogdump(LLV_DEBUG, res->v, res->l);
 
 end:
@@ -967,10 +980,18 @@ oakley_ph1hash_base_i(iph1, sw)
 		return NULL;
 	}
 
-	switch (iph1->approval->authmethod) {
+	switch (AUTHMETHOD(iph1)) {
 	case OAKLEY_ATTR_AUTH_METHOD_PSKEY:
 	case OAKLEY_ATTR_AUTH_METHOD_RSAENC:
 	case OAKLEY_ATTR_AUTH_METHOD_RSAREV:
+#ifdef ENABLE_HYBRID
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_RSAENC_I:
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_RSAENC_R:
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_RSAREV_I:
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_RSAREV_R:
+	case FICTIVE_AUTH_METHOD_XAUTH_PSKEY_I:
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_PSKEY_R:
+#endif
 		if (iph1->skeyid == NULL) {
 			plog(LLV_ERROR, LOCATION, NULL, "no SKEYID found.\n");
 			return NULL;
@@ -980,12 +1001,18 @@ oakley_ph1hash_base_i(iph1, sw)
 
 	case OAKLEY_ATTR_AUTH_METHOD_DSSSIG:
 	case OAKLEY_ATTR_AUTH_METHOD_RSASIG:
+#ifdef HAVE_GSSAPI
 	case OAKLEY_ATTR_AUTH_METHOD_GSSAPI_KRB:
+#endif
 #ifdef ENABLE_HYBRID
-	case OAKLEY_ATTR_AUTH_METHOD_HYBRID_RSA_R:
-	case OAKLEY_ATTR_AUTH_METHOD_HYBRID_DSS_R:
 	case OAKLEY_ATTR_AUTH_METHOD_HYBRID_RSA_I:
+	case OAKLEY_ATTR_AUTH_METHOD_HYBRID_RSA_R:
 	case OAKLEY_ATTR_AUTH_METHOD_HYBRID_DSS_I:
+	case OAKLEY_ATTR_AUTH_METHOD_HYBRID_DSS_R:
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_RSASIG_I:
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_RSASIG_R:
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_DSSSIG_I:
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_DSSSIG_R:
 #endif
 		/* make hash for seed */
 		len = iph1->nonce->l + iph1->nonce_p->l;
@@ -1093,16 +1120,28 @@ oakley_ph1hash_base_r(iph1, sw)
 			"invalid etype for this hash function\n");
 		return NULL;
 	}
-	if (iph1->approval->authmethod != OAKLEY_ATTR_AUTH_METHOD_DSSSIG
+
+	switch(AUTHMETHOD(iph1)) {
+	case OAKLEY_ATTR_AUTH_METHOD_DSSSIG:
+	case OAKLEY_ATTR_AUTH_METHOD_RSASIG:
 #ifdef ENABLE_HYBRID
-	 && iph1->approval->authmethod != OAKLEY_ATTR_AUTH_METHOD_HYBRID_RSA_I
-	 && iph1->approval->authmethod != OAKLEY_ATTR_AUTH_METHOD_HYBRID_DSS_I
+	case OAKLEY_ATTR_AUTH_METHOD_HYBRID_RSA_I:
+	case OAKLEY_ATTR_AUTH_METHOD_HYBRID_RSA_R:
+	case OAKLEY_ATTR_AUTH_METHOD_HYBRID_DSS_I:
+	case OAKLEY_ATTR_AUTH_METHOD_HYBRID_DSS_R:
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_RSASIG_I:
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_RSASIG_R:
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_DSSSIG_I:
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_DSSSIG_R:
+	case FICTIVE_AUTH_METHOD_XAUTH_PSKEY_I:
 #endif
-	 && iph1->approval->authmethod != OAKLEY_ATTR_AUTH_METHOD_RSASIG) {
+		break;
+	default:
 		plog(LLV_ERROR, LOCATION, NULL,
 			"not supported authentication method %d\n",
 			iph1->approval->authmethod);
 		return NULL;
+		break;
 	}
 
 	/* make hash for seed */
@@ -1164,7 +1203,7 @@ oakley_ph1hash_base_r(iph1, sw)
 	memcpy(p, bp->v, bp->l);
 	p += bp->l;
 
-	plog(LLV_DEBUG, LOCATION, NULL, "HASH with:\n");
+	plog(LLV_DEBUG, LOCATION, NULL, "HASH_R with:\n");
 	plogdump(LLV_DEBUG, buf->v, buf->l);
 
 	/* compute HASH */
@@ -1174,7 +1213,7 @@ oakley_ph1hash_base_r(iph1, sw)
 
 	error = 0;
 
-	plog(LLV_DEBUG, LOCATION, NULL, "HASH computed:\n");
+	plog(LLV_DEBUG, LOCATION, NULL, "HASH_R computed:\n");
 	plogdump(LLV_DEBUG, res->v, res->l);
 
 end:
@@ -1209,8 +1248,13 @@ oakley_validate_auth(iph1)
 #ifdef ENABLE_STATS
 	gettimeofday(&start, NULL);
 #endif
-	switch (iph1->approval->authmethod) {
+
+	switch (AUTHMETHOD(iph1)) {
 	case OAKLEY_ATTR_AUTH_METHOD_PSKEY:
+#ifdef ENABLE_HYBRID
+	case FICTIVE_AUTH_METHOD_XAUTH_PSKEY_I:
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_PSKEY_R:
+#endif
 		/* validate HASH */
 	    {
 		char *r_hash;
@@ -1220,10 +1264,20 @@ oakley_validate_auth(iph1)
 				"few isakmp message received.\n");
 			return ISAKMP_NTYPE_PAYLOAD_MALFORMED;
 		}
-
+#ifdef ENABLE_HYBRID
+		if (AUTHMETHOD(iph1) == FICTIVE_AUTH_METHOD_XAUTH_PSKEY_I &&
+		    ((iph1->mode_cfg->flags & ISAKMP_CFG_VENDORID_XAUTH) == 0))
+		{
+			plog(LLV_ERROR, LOCATION, NULL, "No SIG was passed, "
+			    "hybrid auth is enabled, "
+			    "but peer is no Xauth compliant\n");
+			return ISAKMP_NTYPE_SITUATION_NOT_SUPPORTED;
+			break;
+		}
+#endif
 		r_hash = (caddr_t)(iph1->pl_hash + 1);
 
-		plog(LLV_DEBUG, LOCATION, NULL, "HASH received:");
+		plog(LLV_DEBUG, LOCATION, NULL, "HASH received:\n");
 		plogdump(LLV_DEBUG, r_hash,
 			ntohs(iph1->pl_hash->h.len) - sizeof(*iph1->pl_hash));
 
@@ -1260,8 +1314,12 @@ oakley_validate_auth(iph1)
 	case OAKLEY_ATTR_AUTH_METHOD_DSSSIG:
 	case OAKLEY_ATTR_AUTH_METHOD_RSASIG:
 #ifdef ENABLE_HYBRID
-	case OAKLEY_ATTR_AUTH_METHOD_HYBRID_RSA_R:
-	case OAKLEY_ATTR_AUTH_METHOD_HYBRID_DSS_R:
+	case OAKLEY_ATTR_AUTH_METHOD_HYBRID_RSA_I:
+	case OAKLEY_ATTR_AUTH_METHOD_HYBRID_DSS_I:
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_RSASIG_I:
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_RSASIG_R:
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_DSSSIG_I:
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_DSSSIG_R:
 #endif
 	    {
 		int error = 0;
@@ -1354,9 +1412,9 @@ oakley_validate_auth(iph1)
 		 && iph1->rmconf->getcert_method == ISAKMP_GETCERT_PAYLOAD) {
 			certtype = iph1->rmconf->certtype;
 #ifdef ENABLE_HYBRID
-			switch (iph1->approval->authmethod) {
-			case OAKLEY_ATTR_AUTH_METHOD_HYBRID_RSA_R:
-			case OAKLEY_ATTR_AUTH_METHOD_HYBRID_DSS_R:
+			switch (AUTHMETHOD(iph1)) {
+			case OAKLEY_ATTR_AUTH_METHOD_HYBRID_RSA_I:
+			case OAKLEY_ATTR_AUTH_METHOD_HYBRID_DSS_I:
 				certtype = iph1->cert_p->type;
 				break;
 			default:
@@ -1420,9 +1478,9 @@ oakley_validate_auth(iph1)
 
 		certtype = iph1->rmconf->certtype;
 #ifdef ENABLE_HYBRID
-		switch (iph1->approval->authmethod) {
-		case OAKLEY_ATTR_AUTH_METHOD_HYBRID_RSA_R:
-		case OAKLEY_ATTR_AUTH_METHOD_HYBRID_DSS_R:
+		switch (AUTHMETHOD(iph1)) {
+		case OAKLEY_ATTR_AUTH_METHOD_HYBRID_RSA_I:
+		case OAKLEY_ATTR_AUTH_METHOD_HYBRID_DSS_I:
 			certtype = iph1->cert_p->type;
 			break;
 		default:
@@ -1461,8 +1519,8 @@ oakley_validate_auth(iph1)
 	    }
 		break;
 #ifdef ENABLE_HYBRID
-	case OAKLEY_ATTR_AUTH_METHOD_HYBRID_RSA_I:
-	case OAKLEY_ATTR_AUTH_METHOD_HYBRID_DSS_I:
+	case OAKLEY_ATTR_AUTH_METHOD_HYBRID_RSA_R:
+	case OAKLEY_ATTR_AUTH_METHOD_HYBRID_DSS_R:
 	    {
 		if ((iph1->mode_cfg->flags & ISAKMP_CFG_VENDORID_XAUTH) == 0) {
 			plog(LLV_ERROR, LOCATION, NULL, "No SIG was passed, "
@@ -1480,6 +1538,11 @@ oakley_validate_auth(iph1)
 #endif
 #ifdef HAVE_GSSAPI
 	case OAKLEY_ATTR_AUTH_METHOD_GSSAPI_KRB:
+		/* check if we're not into XAUTH_PSKEY_I instead */
+#ifdef ENABLE_HYBRID
+		if (iph1->rmconf->xauth)
+			break;
+#endif
 		switch (iph1->etype) {
 		case ISAKMP_ETYPE_IDENT:
 		case ISAKMP_ETYPE_AGG:
@@ -1517,6 +1580,12 @@ oakley_validate_auth(iph1)
 #endif
 	case OAKLEY_ATTR_AUTH_METHOD_RSAENC:
 	case OAKLEY_ATTR_AUTH_METHOD_RSAREV:
+#ifdef ENABLE_HYBRID
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_RSAENC_I:
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_RSAENC_R:
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_RSAREV_I:
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_RSAREV_R:
+#endif
 		if (iph1->id_p == NULL || iph1->pl_hash == NULL) {
 			plog(LLV_ERROR, LOCATION, iph1->remote,
 				"few isakmp message received.\n");
@@ -1669,28 +1738,39 @@ get_plainrsa_fromlocal(iph1, my)
 	int error = -1;
 
 	iph1->rsa_candidates = rsa_lookup_keys(iph1, my);
-	if (!iph1->rsa_candidates || rsa_list_count(iph1->rsa_candidates) == 0) {
+	if (!iph1->rsa_candidates || 
+	    rsa_list_count(iph1->rsa_candidates) == 0) {
 		plog(LLV_ERROR, LOCATION, NULL,
 			"%s RSA key not found for %s\n",
 			my ? "Private" : "Public",
-			saddr2str_fromto("%s <-> %s", iph1->local, iph1->remote));
+			saddr2str_fromto("%s <-> %s", 
+			iph1->local, iph1->remote));
 		goto end;
 	}
 
 	if (my && rsa_list_count(iph1->rsa_candidates) > 1) {
 		plog(LLV_WARNING, LOCATION, NULL,
-			"More than one (=%lu) private PlainRSA key found for %s\n",
+			"More than one (=%lu) private "
+			"PlainRSA key found for %s\n",
 			rsa_list_count(iph1->rsa_candidates),
-			saddr2str_fromto("%s <-> %s", iph1->local, iph1->remote));
+			saddr2str_fromto("%s <-> %s", 
+			iph1->local, iph1->remote));
 		plog(LLV_WARNING, LOCATION, NULL,
-			"This may have unpredictable results, i.e. wrong key could be used!\n");
+			"This may have unpredictable results, "
+			"i.e. wrong key could be used!\n");
 		plog(LLV_WARNING, LOCATION, NULL,
-			"Consider using only one single private key for all peers...\n");
+			"Consider using only one single private "
+			"key for all peers...\n");
 	}
 	if (my) {
-		iph1->rsa = ((struct rsa_key *)genlist_next(iph1->rsa_candidates, NULL))->rsa;
+		iph1->rsa = ((struct rsa_key *)
+		    genlist_next(iph1->rsa_candidates, NULL))->rsa;
+
 		genlist_free(iph1->rsa_candidates, NULL);
 		iph1->rsa_candidates = NULL;
+
+		if (iph1->rsa == NULL)
+			goto end;
 	}
 
 	error = 0;
@@ -1797,7 +1877,9 @@ oakley_check_certid(iph1)
 		vfree(name);
 		if (error != 0) {
 			plog(LLV_ERROR, LOCATION, NULL,
-				"ID mismatched with subjectAltName.\n");
+				"ID mismatched with ASN1 SubjectName.\n");
+			plogdump(LLV_DEBUG, id_b + 1, idlen);
+			plogdump(LLV_DEBUG, name->v, idlen);
 			return ISAKMP_NTYPE_INVALID_ID_INFORMATION;
 		}
 		return 0;
@@ -1868,6 +1950,8 @@ oakley_check_certid(iph1)
 		if (error != 0) {
 			plog(LLV_ERROR, LOCATION, NULL,
 				"ID mismatched with subjectAltName.\n");
+			plogdump(LLV_DEBUG, id_b + 1, idlen);
+			plogdump(LLV_DEBUG, a, idlen);
 			return ISAKMP_NTYPE_INVALID_ID_INFORMATION;
 		}
 		return 0;
@@ -1916,6 +2000,8 @@ oakley_check_certid(iph1)
 		error = memcmp(id_b + 1, altname, idlen);
 		if (error) {
 			plog(LLV_ERROR, LOCATION, NULL, "ID mismatched.\n");
+			plogdump(LLV_DEBUG, id_b + 1, idlen);
+			plogdump(LLV_DEBUG, altname, idlen);
 			racoon_free(altname);
 			return ISAKMP_NTYPE_INVALID_ID_INFORMATION;
 		}
@@ -2212,6 +2298,12 @@ save_certbuf(gen)
 {
 	cert_t *new;
 
+	if(ntohs(gen->len) <= sizeof(*gen)){
+		plog(LLV_ERROR, LOCATION, NULL,
+			 "Len is too small !!.\n");
+		return NULL;
+	}
+
 	new = oakley_newcert();
 	if (!new) {
 		plog(LLV_ERROR, LOCATION, NULL,
@@ -2287,10 +2379,15 @@ oakley_getcr(iph1)
 			"failed to get cr buffer\n");
 		return NULL;
 	}
-	buf->v[0] = iph1->rmconf->certtype;
-
-	plog(LLV_DEBUG, LOCATION, NULL, "create my CR: %s\n",
+	if(iph1->rmconf->certtype == ISAKMP_CERT_NONE) {
+		buf->v[0] = iph1->rmconf->cacerttype;
+		plog(LLV_DEBUG, LOCATION, NULL, "create my CR: NONE, using %s instead\n",
+		s_isakmp_certtype(iph1->rmconf->cacerttype));
+	} else {
+		buf->v[0] = iph1->rmconf->certtype;
+		plog(LLV_DEBUG, LOCATION, NULL, "create my CR: %s\n",
 		s_isakmp_certtype(iph1->rmconf->certtype));
+	}
 	if (buf->l > 1)
 		plogdump(LLV_DEBUG, buf->v, buf->l);
 
@@ -2334,6 +2431,10 @@ oakley_needcr(type)
 #ifdef ENABLE_HYBRID
 	case OAKLEY_ATTR_AUTH_METHOD_HYBRID_RSA_I:
 	case OAKLEY_ATTR_AUTH_METHOD_HYBRID_DSS_I:
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_RSASIG_I:
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_RSASIG_R:
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_DSSSIG_I:
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_DSSSIG_R:
 #endif
 		return 1;
 	default:
@@ -2357,10 +2458,14 @@ oakley_skeyid(iph1)
 	char *p;
 	int len;
 	int error = -1;
-
+	
 	/* SKEYID */
-	switch(iph1->approval->authmethod) {
+	switch (AUTHMETHOD(iph1)) {
 	case OAKLEY_ATTR_AUTH_METHOD_PSKEY:
+#ifdef ENABLE_HYBRID
+	case FICTIVE_AUTH_METHOD_XAUTH_PSKEY_I:
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_PSKEY_R:
+#endif
 		if (iph1->etype != ISAKMP_ETYPE_IDENT) {
 			iph1->authstr = getpskbyname(iph1->id_p);
 			if (iph1->authstr == NULL) {
@@ -2427,6 +2532,10 @@ oakley_skeyid(iph1)
 	case OAKLEY_ATTR_AUTH_METHOD_HYBRID_DSS_I:
 	case OAKLEY_ATTR_AUTH_METHOD_HYBRID_RSA_R:
 	case OAKLEY_ATTR_AUTH_METHOD_HYBRID_DSS_R:
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_RSASIG_I:
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_RSASIG_R:
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_DSSSIG_I:
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_DSSSIG_R:
 #endif
 #ifdef HAVE_GSSAPI
 	case OAKLEY_ATTR_AUTH_METHOD_GSSAPI_KRB:
@@ -2458,6 +2567,12 @@ oakley_skeyid(iph1)
 		break;
 	case OAKLEY_ATTR_AUTH_METHOD_RSAENC:
 	case OAKLEY_ATTR_AUTH_METHOD_RSAREV:
+#ifdef ENABLE_HYBRID
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_RSAENC_I:
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_RSAENC_R:
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_RSAREV_I:
+	case OAKLEY_ATTR_AUTH_METHOD_XAUTH_RSAREV_R:
+#endif
 		plog(LLV_WARNING, LOCATION, NULL,
 			"not supported authentication method %s\n",
 			s_oakley_attr_method(iph1->approval->authmethod));
@@ -2527,7 +2642,7 @@ oakley_skeyid_dae(iph1)
 	buf = NULL;
 
 	plog(LLV_DEBUG, LOCATION, NULL, "SKEYID_d computed:\n");
-	plogdump(LLV_DEBUG, iph1->skeyid_d->v, iph1->skeyid->l);
+	plogdump(LLV_DEBUG, iph1->skeyid_d->v, iph1->skeyid_d->l);
 
 	/* SKEYID A */
 	/* SKEYID_a = prf(SKEYID, SKEYID_d | g^xy | CKY-I | CKY-R | 1) */
@@ -2939,6 +3054,7 @@ oakley_delivm(ivm)
 	if (ivm->ive != NULL)
 		vfree(ivm->ive);
 	racoon_free(ivm);
+	plog(LLV_DEBUG, LOCATION, NULL, "IV freed\n");
 
 	return;
 }
@@ -3003,8 +3119,6 @@ oakley_do_decrypt(iph1, msg, ivdp, ivep)
 
 	vfree(buf);
 	buf = NULL;
-	if (new == NULL)
-		goto end;
 
 	plog(LLV_DEBUG, LOCATION, NULL, "decrypted payload by IV:\n");
 	plogdump(LLV_DEBUG, ivdp->v, ivdp->l);
@@ -3138,8 +3252,6 @@ oakley_do_encrypt(iph1, msg, ivep, ivp)
 
 	vfree(buf);
 	buf = NULL;
-	if (new == NULL)
-		goto end;
 
 	plog(LLV_DEBUG, LOCATION, NULL, "encrypted payload by IV:\n");
 	plogdump(LLV_DEBUG, ivep->v, ivep->l);
